@@ -1,19 +1,31 @@
-import unittest
 import io
-import imghdr
+import os
+import unittest
+from unittest import mock
 
+import filetype
 from PIL import Image as PILImage
+from PIL import ImageChops
 
-from willow.image import JPEGImageFile, PNGImageFile, GIFImageFile, WebPImageFile
-from willow.plugins.pillow import _PIL_Image, PillowImage, UnsupportedRotation
-
+from willow.image import (
+    AvifImageFile,
+    BadImageOperationError,
+    GIFImageFile,
+    JPEGImageFile,
+    PNGImageFile,
+    WebPImageFile,
+)
+from willow.optimizers import Cwebp, Gifsicle, Jpegoptim, Optipng, Pngquant
+from willow.plugins.pillow import PillowImage, UnsupportedRotation, _PIL_Image
+from willow.registry import registry
 
 no_webp_support = not PillowImage.is_format_supported("WEBP")
+no_avif_support = not PillowImage.is_format_supported("AVIF")
 
 
 class TestPillowOperations(unittest.TestCase):
     def setUp(self):
-        with open('tests/images/transparent.png', 'rb') as f:
+        with open("tests/images/transparent.png", "rb") as f:
             self.image = PillowImage.open(PNGImageFile(f))
 
     def test_get_size(self):
@@ -33,14 +45,60 @@ class TestPillowOperations(unittest.TestCase):
         cropped_image = self.image.crop((10, 10, 100, 100))
         self.assertEqual(cropped_image.get_size(), (90, 90))
 
+    def test_crop_out_of_bounds(self):
+        # crop rectangle should be clamped to the image boundaries
+        bottom_right_cropped_image = self.image.crop((150, 100, 250, 200))
+        self.assertEqual(bottom_right_cropped_image.get_size(), (50, 50))
+
+        top_left_cropped_image = self.image.crop((-50, -50, 50, 50))
+        self.assertEqual(top_left_cropped_image.get_size(), (50, 50))
+
+        # fail if the crop rectangle is entirely to the left of the image
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((-100, 50, -50, 100))
+        # right edge of crop rectangle is exclusive, so 0 is also invalid
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((-50, 50, 0, 100))
+
+        # fail if the crop rectangle is entirely above the image
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((50, -100, 100, -50))
+        # bottom edge of crop rectangle is exclusive, so 0 is also invalid
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((50, -50, 100, 0))
+
+        # fail if the crop rectangle is entirely to the right of the image
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((250, 50, 300, 100))
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((200, 50, 250, 100))
+
+        # fail if the crop rectangle is entirely below the image
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((50, 200, 100, 250))
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((50, 150, 100, 200))
+
+        # fail if left edge >= right edge
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((125, 25, 25, 125))
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((100, 25, 100, 125))
+
+        # fail if bottom edge >= top edge
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((25, 125, 125, 25))
+        with self.assertRaises(BadImageOperationError):
+            self.image.crop((25, 100, 125, 100))
+
     def test_rotate(self):
         rotated_image = self.image.rotate(90)
         width, height = rotated_image.get_size()
         self.assertEqual((width, height), (150, 200))
 
     def test_rotate_without_multiple_of_90(self):
-        with self.assertRaises(UnsupportedRotation) as e:
-            rotated_image = self.image.rotate(45)
+        with self.assertRaises(UnsupportedRotation):
+            self.image.rotate(45)
 
     def test_rotate_greater_than_360(self):
         # 450 should end up the same as a 90 rotation
@@ -60,9 +118,11 @@ class TestPillowOperations(unittest.TestCase):
 
     def test_set_background_color_rgb_color_argument_check(self):
         with self.assertRaises(TypeError) as e:
-            self.image.set_background_color_rgb('rgb(255, 0, 0)')
+            self.image.set_background_color_rgb("rgb(255, 0, 0)")
 
-        self.assertEqual(str(e.exception), "the 'color' argument must be a 3-element tuple or list")
+        self.assertEqual(
+            str(e.exception), "the 'color' argument must be a 3-element tuple or list"
+        )
 
     def test_save_as_jpeg(self):
         # Remove alpha channel from image
@@ -72,7 +132,7 @@ class TestPillowOperations(unittest.TestCase):
         return_value = image.save_as_jpeg(output)
         output.seek(0)
 
-        self.assertEqual(imghdr.what(output), 'jpeg')
+        self.assertEqual(filetype.guess_extension(output), "jpg")
         self.assertIsInstance(return_value, JPEGImageFile)
         self.assertEqual(return_value.f, output)
 
@@ -92,14 +152,14 @@ class TestPillowOperations(unittest.TestCase):
 
         image = image.save_as_jpeg(io.BytesIO(), progressive=True)
 
-        self.assertTrue(PILImage.open(image.f).info['progressive'])
+        self.assertTrue(PILImage.open(image.f).info["progressive"])
 
     def test_save_as_png(self):
         output = io.BytesIO()
         return_value = self.image.save_as_png(output)
         output.seek(0)
 
-        self.assertEqual(imghdr.what(output), 'png')
+        self.assertEqual(filetype.guess_extension(output), "png")
         self.assertIsInstance(return_value, PNGImageFile)
         self.assertEqual(return_value.f, output)
 
@@ -110,27 +170,42 @@ class TestPillowOperations(unittest.TestCase):
         # Optimised image must be smaller than unoptimised image
         self.assertTrue(optimised.f.tell() < unoptimised.f.tell())
 
+    def test_save_mode_cmyk_as_png(self):
+        output = io.BytesIO()
+
+        with open("tests/images/cmyk.jpg", "rb") as f:
+            image = PillowImage.open(JPEGImageFile(f))
+
+        return_value = image.save_as_png(output)
+        output.seek(0)
+
+        image = _PIL_Image().open(output)
+        self.assertEqual(image.mode, "RGB")
+        self.assertEqual(filetype.guess_extension(output), "png")
+        self.assertIsInstance(return_value, PNGImageFile)
+        self.assertEqual(return_value.f, output)
+
     def test_save_as_gif(self):
         output = io.BytesIO()
         return_value = self.image.save_as_gif(output)
         output.seek(0)
 
-        self.assertEqual(imghdr.what(output), 'gif')
+        self.assertEqual(filetype.guess_extension(output), "gif")
         self.assertIsInstance(return_value, GIFImageFile)
         self.assertEqual(return_value.f, output)
 
     def test_save_as_gif_converts_back_to_supported_mode(self):
         output = io.BytesIO()
 
-        with open('tests/images/transparent.gif', 'rb') as f:
+        with open("tests/images/transparent.gif", "rb") as f:
             image = PillowImage.open(GIFImageFile(f))
-            image.image = image.image.convert('RGB')
+            image.image = image.image.convert("RGB")
 
         image.save_as_gif(output)
         output.seek(0)
 
         image = _PIL_Image().open(output)
-        self.assertEqual(image.mode, 'P')
+        self.assertEqual(image.mode, "P")
 
     def test_has_alpha(self):
         has_alpha = self.image.has_alpha()
@@ -141,17 +216,17 @@ class TestPillowOperations(unittest.TestCase):
         self.assertFalse(has_animation)
 
     def test_transparent_gif(self):
-        with open('tests/images/transparent.gif', 'rb') as f:
+        with open("tests/images/transparent.gif", "rb") as f:
             image = PillowImage.open(GIFImageFile(f))
 
         self.assertTrue(image.has_alpha())
         self.assertFalse(image.has_animation())
 
         # Check that the alpha of pixel 1,1 is 0
-        self.assertEqual(image.image.convert('RGBA').getpixel((1, 1))[3], 0)
+        self.assertEqual(image.image.convert("RGBA").getpixel((1, 1))[3], 0)
 
     def test_resize_transparent_gif(self):
-        with open('tests/images/transparent.gif', 'rb') as f:
+        with open("tests/images/transparent.gif", "rb") as f:
             image = PillowImage.open(GIFImageFile(f))
 
         resized_image = image.resize((100, 75))
@@ -160,10 +235,10 @@ class TestPillowOperations(unittest.TestCase):
         self.assertFalse(resized_image.has_animation())
 
         # Check that the alpha of pixel 1,1 is 0
-        self.assertEqual(resized_image.image.convert('RGBA').getpixel((1, 1))[3], 0)
+        self.assertEqual(resized_image.image.convert("RGBA").getpixel((1, 1))[3], 0)
 
     def test_save_transparent_gif(self):
-        with open('tests/images/transparent.gif', 'rb') as f:
+        with open("tests/images/transparent.gif", "rb") as f:
             image = PillowImage.open(GIFImageFile(f))
 
         # Save it into memory
@@ -178,11 +253,11 @@ class TestPillowOperations(unittest.TestCase):
         self.assertFalse(image.has_animation())
 
         # Check that the alpha of pixel 1,1 is 0
-        self.assertEqual(image.image.convert('RGBA').getpixel((1, 1))[3], 0)
+        self.assertEqual(image.image.convert("RGBA").getpixel((1, 1))[3], 0)
 
     @unittest.expectedFailure  # Pillow doesn't support animation
     def test_animated_gif(self):
-        with open('tests/images/newtons_cradle.gif', 'rb') as f:
+        with open("tests/images/newtons_cradle.gif", "rb") as f:
             image = PillowImage.open(GIFImageFile(f))
 
         self.assertFalse(image.has_alpha())
@@ -190,7 +265,7 @@ class TestPillowOperations(unittest.TestCase):
 
     @unittest.expectedFailure  # Pillow doesn't support animation
     def test_resize_animated_gif(self):
-        with open('tests/images/newtons_cradle.gif', 'rb') as f:
+        with open("tests/images/newtons_cradle.gif", "rb") as f:
             image = PillowImage.open(GIFImageFile(f))
 
         resized_image = image.resize((100, 75))
@@ -209,13 +284,13 @@ class TestPillowOperations(unittest.TestCase):
         return_value = self.image.save_as_webp(output)
         output.seek(0)
 
-        self.assertEqual(imghdr.what(output), 'webp')
+        self.assertEqual(filetype.guess_extension(output), "webp")
         self.assertIsInstance(return_value, WebPImageFile)
         self.assertEqual(return_value.f, output)
 
     @unittest.skipIf(no_webp_support, "Pillow does not have WebP support")
     def test_open_webp(self):
-        with open('tests/images/tree.webp', 'rb') as f:
+        with open("tests/images/tree.webp", "rb") as f:
             image = PillowImage.open(WebPImageFile(f))
 
         self.assertFalse(image.has_alpha())
@@ -223,34 +298,133 @@ class TestPillowOperations(unittest.TestCase):
 
     @unittest.skipIf(no_webp_support, "Pillow does not have WebP support")
     def test_open_webp_w_alpha(self):
-        with open('tests/images/tux_w_alpha.webp', 'rb') as f:
+        with open("tests/images/tux_w_alpha.webp", "rb") as f:
             image = PillowImage.open(WebPImageFile(f))
 
         self.assertTrue(image.has_alpha())
         self.assertFalse(image.has_animation())
 
     @unittest.skipIf(no_webp_support, "Pillow does not have WebP support")
-    def test_open_webp_quality(self):
+    def test_save_webp_quality(self):
         high_quality = self.image.save_as_webp(io.BytesIO(), quality=90)
         low_quality = self.image.save_as_webp(io.BytesIO(), quality=30)
         self.assertTrue(low_quality.f.tell() < high_quality.f.tell())
 
     @unittest.skipIf(no_webp_support, "Pillow does not have WebP support")
-    def test_open_webp_lossless(self):
+    def test_save_webp_lossless(self):
         original_image = self.image.image
+
         lossless_file = self.image.save_as_webp(io.BytesIO(), lossless=True)
         lossless_image = PillowImage.open(lossless_file).image
-        identically = True
-        for x in range(original_image.width):
-            for y in range(original_image.height):
-                original_pixel = original_image.getpixel((x, y))
-                # don't compare fully transparent pixels
-                if original_pixel[3] == 0:
-                    continue
-                if original_pixel != lossless_image.getpixel((x, y)):
-                    identically = False
-                    break
-        self.assertTrue(identically)
+
+        diff = ImageChops.difference(original_image, lossless_image)
+        self.assertIsNone(diff.getbbox())
+
+    @unittest.skipIf(no_avif_support, "Pillow does not have AVIF support")
+    def test_save_as_avif(self):
+        output = io.BytesIO()
+        return_value = self.image.save_as_avif(output)
+        output.seek(0)
+
+        self.assertEqual(filetype.guess_extension(output), "avif")
+        self.assertIsInstance(return_value, AvifImageFile)
+        self.assertEqual(return_value.f, output)
+
+    @unittest.skipIf(no_avif_support, "Pillow does not have AVIF support")
+    def test_open_avif(self):
+        with open("tests/images/tree.webp", "rb") as f:
+            image = PillowImage.open(AvifImageFile(f))
+
+        self.assertFalse(image.has_alpha())
+        self.assertFalse(image.has_animation())
+
+    @unittest.skipIf(no_avif_support, "Pillow does not have AVIF support")
+    def test_save_avif_quality(self):
+        high_quality = self.image.save_as_avif(io.BytesIO(), quality=90)
+        low_quality = self.image.save_as_avif(io.BytesIO(), quality=30)
+        self.assertTrue(low_quality.f.tell() < high_quality.f.tell())
+
+    @unittest.skipIf(no_avif_support, "Pillow does not have AVIF support")
+    def test_save_avif_lossless(self):
+        original_image = self.image.image
+
+        lossless_file = self.image.save_as_avif(io.BytesIO(), lossless=True)
+        lossless_image = PillowImage.open(lossless_file).image
+
+        diff = ImageChops.difference(original_image, lossless_image)
+        self.assertIsNone(diff.getbbox())
+
+
+class TestPillowImageWithOptimizers(unittest.TestCase):
+    def setUp(self):
+        with mock.patch.dict(os.environ, {"WILLOW_OPTIMIZERS": "true"}):
+            registry.register_optimizer(Cwebp)
+            registry.register_optimizer(Gifsicle)
+            registry.register_optimizer(Jpegoptim)
+            registry.register_optimizer(Optipng)
+            registry.register_optimizer(Pngquant)
+
+    def tearDown(self):
+        # reset the registry as we get the global state
+        registry._registered_optimizers = []
+
+    @unittest.skipIf(not Jpegoptim.check_library(), "jpegoptim not installed")
+    def test_save_as_jpeg(self):
+        with open("tests/images/flower.jpg", "rb") as f:
+            original_size = os.fstat(f.fileno()).st_size
+            image = PillowImage.open(JPEGImageFile(f))
+
+        return_value = image.save_as_jpeg(io.BytesIO())
+        self.assertTrue(original_size > return_value.f.seek(0, io.SEEK_END))
+
+        with mock.patch("willow.plugins.pillow.PillowImage.optimize") as mock_optimize:
+            image.save_as_jpeg(io.BytesIO(), apply_optimizers=False)
+            mock_optimize.assert_not_called()
+
+    @unittest.skipIf(
+        not (Pngquant.check_library() and Optipng.check_library()),
+        "optipng or pngquant not installed",
+    )
+    def test_save_as_png(self):
+        with open("tests/images/transparent.png", "rb") as f:
+            original_size = os.fstat(f.fileno()).st_size
+            image = PillowImage.open(PNGImageFile(f))
+
+        return_value = image.save_as_png(io.BytesIO())
+        self.assertTrue(original_size > return_value.f.seek(0, io.SEEK_END))
+
+        with mock.patch("willow.plugins.pillow.PillowImage.optimize") as mock_optimize:
+            image.save_as_png(io.BytesIO(), apply_optimizers=False)
+            mock_optimize.assert_not_called()
+
+    @unittest.skipIf(not Gifsicle.check_library(), "gifsicle not installed")
+    def test_save_as_gif(self):
+        with open("tests/images/transparent.gif", "rb") as f:
+            original_size = f.tell()
+            image = PillowImage.open(GIFImageFile(f))
+
+        return_value = image.save_as_gif(io.BytesIO())
+        self.assertTrue(original_size < return_value.f.tell())
+
+        with mock.patch("willow.plugins.pillow.PillowImage.optimize") as mock_optimize:
+            image.save_as_gif(io.BytesIO(), apply_optimizers=False)
+            mock_optimize.assert_not_called()
+
+    @unittest.skipIf(
+        no_webp_support or not Cwebp.check_library(),
+        "webp not supported or cwebp not installed",
+    )
+    def test_save_as_webp(self):
+        with open("tests/images/tree.webp", "rb") as f:
+            original_size = os.fstat(f.fileno()).st_size
+            image = PillowImage.open(WebPImageFile(f))
+
+        return_value = image.save_as_gif(io.BytesIO())
+        self.assertTrue(original_size < return_value.f.tell())
+
+        with mock.patch("willow.plugins.pillow.PillowImage.optimize") as mock_optimize:
+            image.save_as_webp(io.BytesIO(), apply_optimizers=False)
+            mock_optimize.assert_not_called()
 
 
 class TestPillowImageOrientation(unittest.TestCase):
@@ -260,19 +434,19 @@ class TestPillowImageOrientation(unittest.TestCase):
 
         # Check that the red flower is in the bottom left
         # The JPEGs have compressed slightly differently so the colours won't be spot on
-        colour = image.image.convert('RGB').getpixel((155, 282))
+        colour = image.image.convert("RGB").getpixel((155, 282))
         self.assertAlmostEqual(colour[0], 217, delta=25)
         self.assertAlmostEqual(colour[1], 38, delta=25)
         self.assertAlmostEqual(colour[2], 46, delta=25)
 
         # Check that the water is at the bottom
-        colour = image.image.convert('RGB').getpixel((377, 434))
+        colour = image.image.convert("RGB").getpixel((377, 434))
         self.assertAlmostEqual(colour[0], 85, delta=25)
         self.assertAlmostEqual(colour[1], 93, delta=25)
         self.assertAlmostEqual(colour[2], 65, delta=25)
 
     def test_jpeg_with_orientation_1(self):
-        with open('tests/images/orientation/landscape_1.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_1.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -280,7 +454,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_2(self):
-        with open('tests/images/orientation/landscape_2.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_2.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -288,7 +462,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_3(self):
-        with open('tests/images/orientation/landscape_3.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_3.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -296,7 +470,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_4(self):
-        with open('tests/images/orientation/landscape_4.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_4.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -304,7 +478,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_5(self):
-        with open('tests/images/orientation/landscape_5.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_5.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -312,7 +486,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_6(self):
-        with open('tests/images/orientation/landscape_6.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_6.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -320,7 +494,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_7(self):
-        with open('tests/images/orientation/landscape_7.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_7.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
@@ -328,7 +502,7 @@ class TestPillowImageOrientation(unittest.TestCase):
         self.assert_orientation_landscape_image_is_correct(image)
 
     def test_jpeg_with_orientation_8(self):
-        with open('tests/images/orientation/landscape_8.jpg', 'rb') as f:
+        with open("tests/images/orientation/landscape_8.jpg", "rb") as f:
             image = PillowImage.open(JPEGImageFile(f))
 
         image = image.auto_orient()
